#include "postgres.h"

#include "access/hash.h"
#include "fmgr.h"
#include "funcapi.h"
#include "executor/spi.h"
#include "miscadmin.h"
#include "utils/builtins.h"
#include "utils/formatting.h"
#include "utils/inet.h"

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif

extern Datum inet_test(PG_FUNCTION_ARGS);

static bool
compatTupleDescs(TupleDesc ret_tupdesc, TupleDesc sql_tupdesc)
{
	int			i;
	Form_pg_attribute	ret_attr;
	Form_pg_attribute	sql_attr;

	for (i = 0; i < ret_tupdesc->natts; i++)
	{
		ret_attr = ret_tupdesc->attrs[i];
		sql_attr = sql_tupdesc->attrs[i];
		if (ret_attr->atttypid != sql_attr->atttypid)
			return false;
	}

	return true;
}

PG_FUNCTION_INFO_V1(inet_test);

Datum
inet_test(PG_FUNCTION_ARGS)
{
	ReturnSetInfo	   *rsinfo = (ReturnSetInfo *) fcinfo->resultinfo;
	Tuplestorestate	   *tupstore;
	TupleDesc		tupdesc;
	int			call_cntr;
	int			max_calls;
	SPITupleTable	   *spi_tuptable;
	TupleDesc		spi_tupdesc;
	MemoryContext		per_query_ctx;
	MemoryContext		oldcontext;
	int			ret;
	int			proc;
	Datum			values[3];
	bool			isnull[3];

	/* check to see if caller supports us returning a tuplestore */
	if (rsinfo == NULL || !IsA(rsinfo, ReturnSetInfo))
		ereport(ERROR,
				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
				 errmsg("set-valued function called in context that cannot accept a set")));
	if (!(rsinfo->allowedModes & SFRM_Materialize))
		ereport(ERROR,
				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
				 errmsg("materialize mode required, but it is not " \
				 	"allowed in this context")));

	per_query_ctx = rsinfo->econtext->ecxt_per_query_memory;
	/* Connect to SPI manager */
	if ((ret = SPI_connect()) < 0)
		/* internal error */
		elog(ERROR, "inet_test: SPI_connect returned %d", ret);

	/* Retrieve the desired rows */
	ret = SPI_execute("select t1.id, t1.i1, t2.i2 from t1 join t2 on (t1.id = t2.id2);", true, 0);
	proc = SPI_processed;

	/* If no qualifying tuples, fall out early */
	if (ret != SPI_OK_SELECT || proc <= 0)
	{
		SPI_finish();
		rsinfo->isDone = ExprEndResult;
		PG_RETURN_NULL();
	}

	spi_tuptable = SPI_tuptable;
	spi_tupdesc = spi_tuptable->tupdesc;

	/* get a tuple descriptor for our result type */
	switch (get_call_result_type(fcinfo, NULL, &tupdesc))
	{
		case TYPEFUNC_COMPOSITE:
			/* success */
			break;
		case TYPEFUNC_RECORD:
			/* failed to determine actual type of RECORD */
			ereport(ERROR,
					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
					 errmsg("function returning record called in context "
					 	"that cannot accept type record")));
			break;
		default:
			/* result type isn't composite */
			elog(ERROR, "return type must be a row type");
			break;
	}

	/*
	 * Check that return tupdesc is compatible with the data we got from SPI,
	 * at least based on number and type of attributes
	 */
	if (!compatTupleDescs(tupdesc, spi_tupdesc))
		ereport(ERROR,
				(errcode(ERRCODE_SYNTAX_ERROR),
				 errmsg("return and sql tuple descriptions are " \
				 	"incompatible")));

	/*
	 * switch to long-lived memory context
	 */
	oldcontext = MemoryContextSwitchTo(per_query_ctx);

	/* make sure we have a persistent copy of the result tupdesc */
	tupdesc = CreateTupleDescCopy(tupdesc);

	/* initialize our tuplestore in long-lived context */
	tupstore =
			tuplestore_begin_heap(rsinfo->allowedModes & SFRM_Materialize_Random,
									false, work_mem);

	MemoryContextSwitchTo(oldcontext);

	/* total number of tuples to be examined */
	max_calls = proc;

	/* the return tuple always must have 1 rowid + num_categories columns */

	for (call_cntr = 0; call_cntr < max_calls; call_cntr++)
	{
		HeapTuple	spi_tuple;
		HeapTuple	tuple;

		/* get the next sql result tuple */
		spi_tuple = spi_tuptable->vals[call_cntr];

		values[0] = SPI_getbinval(spi_tuple, spi_tupdesc, 1, &isnull[0]);
		values[1] = SPI_getbinval(spi_tuple, spi_tupdesc, 2, &isnull[1]);
		values[2] = SPI_getbinval(spi_tuple, spi_tupdesc, 3, &isnull[2]);

		/* Test INET values  */
		if (!isnull[1])
		{
#if 0
			inet	*i1 = DatumGetInetP(values[1]);
#else
			inet	*i1 = (inet *)PG_DETOAST_DATUM(values[1]);
#endif

			if (i1->inet_data.family == PGSQL_AF_INET)
			{
				elog(NOTICE, "i1 family=AF_INET");
				elog(NOTICE, "i1 netmask=%d bits", (int)i1->inet_data.bits);
				elog(NOTICE, "i1 address=%d.%d.%d.%d",
							(int)i1->inet_data.ipaddr[0],
							(int)i1->inet_data.ipaddr[1],
							(int)i1->inet_data.ipaddr[2],
							(int)i1->inet_data.ipaddr[3]);
			}
			else if (i1->inet_data.family == PGSQL_AF_INET6)
			{
				elog(NOTICE, "i1 family=AF_INET6");
				elog(NOTICE, "i1 netmask=%d bits", (int)i1->inet_data.bits);
				elog(NOTICE, "i1 address=%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x",
							i1->inet_data.ipaddr[ 0],
							i1->inet_data.ipaddr[ 1],
							i1->inet_data.ipaddr[ 2],
							i1->inet_data.ipaddr[ 3],
							i1->inet_data.ipaddr[ 4],
							i1->inet_data.ipaddr[ 5],
							i1->inet_data.ipaddr[ 6],
							i1->inet_data.ipaddr[ 7],
							i1->inet_data.ipaddr[ 8],
							i1->inet_data.ipaddr[ 9],
							i1->inet_data.ipaddr[10],
							i1->inet_data.ipaddr[11],
							i1->inet_data.ipaddr[12],
							i1->inet_data.ipaddr[13],
							i1->inet_data.ipaddr[14],
							i1->inet_data.ipaddr[15]);
							
			}
			else
				elog(NOTICE, "i1 family=CORRUPTED");
		}

		/* build the tuple and store it */
		tuple = heap_form_tuple(tupdesc, values, isnull);
		tuplestore_puttuple(tupstore, tuple);
		heap_freetuple(tuple);

	}

	/* let the caller know we're sending back a tuplestore */
	rsinfo->returnMode = SFRM_Materialize;
	rsinfo->setResult = tupstore;
	rsinfo->setDesc = tupdesc;

	/* release SPI related resources (and return to caller's context) */
	SPI_finish();

	return (Datum) 0;
}
